"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging
from _socket import SOL_SOCKET, SOL_TCP, SO_KEEPALIVE, TCP_KEEPIDLE, TCP_KEEPINTVL, TCP_KEEPCNT
from _ssl import PROTOCOL_TLSv1
from threading import Lock

import gevent
import socks
from gevent import GreenletExit
from gevent._sslgte279 import SSLSocket
from gevent.queue import Queue
from pysolbase.SolBase import SolBase
from pysolmeters.AtomicInt import AtomicIntSafe

from pysoltcp.tcpbase.ProtocolParserTextDelimited import ProtocolParserTextDelimited
from pysoltcp.tcpbase.TcpSocketManager import TcpSocketManager
from pysoltcp.tcpclient.TcpClientConfig import TcpClientConfig

SolBase.voodoo_init()

logger = logging.getLogger(__name__)


class TcpSimpleClient(TcpSocketManager):
    """
    Tcp simple client
    """

    def __init__(self, tcp_client_config):
        """
        Constructor.
        Option via TcpClientConfig to enable SSL on socket.
        In this case, SSL handshake is to be done manually AFTER connect. Higher level implementation MUST handle this.
        :param tcp_client_config: The tcp client config.
        :return Nothing.
        """

        # Base - we provide two callback :
        # - one for disconnecting ourselves
        # - one to notify socket receive buffer
        TcpSocketManager.__init__(self, self.disconnect, self._on_receive)

        # Check
        if tcp_client_config is None:
            logger.error("TcpSimpleClient : tcp_serverConfig is None")
            raise Exception("TcpSimpleClient : tcp_serverConfig is None")
        elif not isinstance(tcp_client_config, TcpClientConfig):
            logger.error("TcpSimpleClient : tcp_serverConfig is not a TcpServerConfig, class=%s", SolBase.get_classname(tcp_client_config))
            raise Exception("TcpSimpleClient : tcp_serverConfig is not a TcpServerConfig")

        # Default
        self._tcp_client_config = tcp_client_config

        # Greenlets
        self._read_greenlet = None
        self._write_greenlet = None

        # Receive queue
        self._receive_queue = Queue()

        # Receive current buffer
        self._receive_current_buf = None

        # Socket info
        self._local_addr = None
        self._local_port = None
        self._remote_addr = None
        self._remote_port = None
        self._connect_count = AtomicIntSafe()
        self._disconnect_count = AtomicIntSafe()

        # Pool stuff
        self.process_id = None
        self.pool_id = None
        self.dt_alloc = SolBase.datecurrent()
        self.dt_last_acquire = self.dt_alloc
        self.dt_last_release = self.dt_alloc
        self.count_acquire = 0
        self.count_release = 0

        # Pool disconnect notify
        self.on_disconnect_notifypoolcallbacklock = Lock()
        self.on_disconnect_notifypoolcallbackcall = True
        self.on_disconnect_notifypoolcallback = None

    # ================================
    # POOL METHOD
    # ================================

    def notify_acquire(self):
        """
        Notify
        """
        self.dt_last_acquire = SolBase.datecurrent()
        self.count_acquire += 1

    def notify_release(self):
        """
        Notify
        """
        self.dt_last_release = SolBase.datecurrent()
        self.count_release += 1

    # ================================
    # TO STRING OVERWRITE
    # ================================

    def __str__(self):
        """
        To string override
        :return: A string
        :rtype string
        """

        return "c.addr={0}:{1}/{2}:{3}*cc={4}*dc={5}*c.q.recv.size={6}*sock={7}*proxy={8}/{9}:{10}*{11}".format(
            self._local_addr, self._local_port, self._remote_addr, self._remote_port,
            self._connect_count.get(), self._disconnect_count.get(),
            self._receive_queue.qsize(),
            self.current_socket,
            self._tcp_client_config.proxy_enable,
            self._tcp_client_config.proxy_addr,
            self._tcp_client_config.proxy_port,
            TcpSocketManager.__str__(self)
        )

    # =================================
    # CONNECT
    # =================================

    def connect(self):
        """
        Connect to server. Return true upon success.
        If SSL is enable, do NOT forget at higher level to initiate a SSL handshake manually.
        :return True if connected, false otherwise.
        """

        try:
            logger.debug("TcpSimpleClient : connect : starting, target=%s:%s", self._tcp_client_config.target_addr, self._tcp_client_config.target_port)

            # Reset the error flag right now
            self.unset_internal_fatal_error_for_reconnect()

            # Check
            if self.is_connected:
                logger.warn("TcpSimpleClient : connect : already connected, doing nothing, self=%s", self)
                return False

            # Init
            self._connect_count.increment()
            self._dt_created = SolBase.datecurrent()
            self._dt_last_recv = self._dt_created
            self._dt_last_send = self._dt_created

            # ------------------------
            # ALLOC
            # ------------------------
            if self._tcp_client_config.ssl_enable:
                if self._tcp_client_config.proxy_enable:
                    # ------------------------
                    # PROXY ON, SSL ON
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY ON/SSL ON, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY ON/SSL ON, self=%s", self)

                    # Alloc
                    self.current_socket = socks.socksocket()

                    # Proxy
                    self.current_socket.setproxy(socks.PROXY_TYPE_SOCKS5, self._tcp_client_config.proxy_addr, self._tcp_client_config.proxy_port)

                    # SSL Wrap : Inside POST CONNECT
                else:
                    # ------------------------
                    # PROXY OFF, SSL ON
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY OFF/SSL ON, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY OFF/SSL ON, self=%s", self)

                    # Alloc
                    self.current_socket = gevent.socket.socket()

                    # Wrap
                    self.current_socket = SSLSocket(self.current_socket, do_handshake_on_connect=False, ssl_version=PROTOCOL_TLSv1)
            else:
                if self._tcp_client_config.proxy_enable:
                    # ------------------------
                    # PROXY ON, SSL OFF
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY ON/SSL OFF, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY ON/SSL OFF, self=%s", self)

                    # Alloc
                    self.current_socket = socks.socksocket()

                    # Proxy
                    self.current_socket.setproxy(socks.PROXY_TYPE_SOCKS5, self._tcp_client_config.proxy_addr,
                                                 self._tcp_client_config.proxy_port)
                else:
                    # ------------------------
                    # PROXY OFF, SSL OFF
                    # ------------------------
                    if self._tcp_client_config.debug_log:
                        logger.info("TcpSimpleClient : Alloc : PROXY OFF/SSL OFF, self=%s", self)
                    else:
                        logger.debug("TcpSimpleClient : Alloc : PROXY OFF/SSL OFF, self=%s", self)

                    # Alloc
                    self.current_socket = gevent.socket.socket()

            # ------------------------
            # POST ALLOC
            # ------------------------
            if self._tcp_client_config.timeout_ms:
                logger.debug("TcpSimpleClient : connect : setting timeout=%s, self=%s", self._tcp_client_config.timeout_ms, self)
                self.current_socket.settimeout(None)
                logger.debug("TcpSimpleClient : connect : timeout=%s, self=%s", self.current_socket.gettimeout(), self)

            # ------------------------
            # CONNECT
            # ------------------------

            # 2-tuple (host, port)
            address = (self._tcp_client_config.target_addr, self._tcp_client_config.target_port)

            # Connect
            if self._tcp_client_config.debug_log:
                logger.info("TcpSimpleClient : connect : starting, ssl=%s, timeout=%s, self=%s", self._tcp_client_config.ssl_enable, self._tcp_client_config.timeout_ms, self)

            # Go
            logger.debug("TcpSimpleClient : connect : calling connect, self=%s", self)
            self.current_socket.connect(address)
            logger.debug("TcpSimpleClient : connect : now connected, self=%s", self)

            # Store
            self._local_addr = self.current_socket.getsockname()[0]
            self._local_port = self.current_socket.getsockname()[1]
            self._remote_addr = self.current_socket.getpeername()[0]
            self._remote_port = self.current_socket.getpeername()[1]

            # Log
            if self._tcp_client_config.debug_log:
                logger.info("TcpSimpleClient : connected, %s:%s to %s:%s, self=%s", self._local_addr, self._local_port, self._remote_addr, self._remote_port, self)

            # Done
            self.is_connected = True
            self.on_disconnect_notifypoolcallbackcall = True

            # ------------------------
            # POST CONNECT
            # ------------------------
            if self._tcp_client_config.ssl_enable:
                # Switch to SSL now
                if self._tcp_client_config.debug_log:
                    logger.info("PROXY ON/SSL ON, switching to SSL now, self=%s", self)
                else:
                    logger.debug("PROXY ON/SSL ON, switching to SSL now, self=%s", self)

                # Wrap (already connected
                self.current_socket = SSLSocket(self.current_socket, do_handshake_on_connect=False, ssl_version=PROTOCOL_TLSv1)

                # Start
                dt_start = SolBase.datecurrent()

                # Do it
                logger.debug("TcpSimpleClient : __do_ssl_handshake now, self=%s", self)
                self.__do_ssl_handshake()

                # Time
                self._set_ssl_handshake_ms(SolBase.datediff(dt_start))

            # ------------------------
            # POST CONNECT
            # ------------------------

            if self._tcp_client_config.tcp_keepalive_enabled:
                # Switch to TCP KA now
                if self._tcp_client_config.debug_log:
                    logger.info(
                        "TCP KA ON, switching to KA now, (on=%s/delay=%s/failed=%s/interval=%s), self=%s",
                        self._tcp_client_config.tcp_keepalive_enabled,
                        self._tcp_client_config.tcp_keepalive_probes_senddelayms,
                        self._tcp_client_config.tcp_keepalive_probes_failedcount,
                        self._tcp_client_config.tcp_keepalive_probes_sendintervalms,
                        self)
                else:
                    logger.debug(
                        "TCP KA ON, switching to KA now, (on=%s/delay=%s/failed=%s/interval=%s), self=%s",
                        self._tcp_client_config.tcp_keepalive_enabled,
                        self._tcp_client_config.tcp_keepalive_probes_senddelayms,
                        self._tcp_client_config.tcp_keepalive_probes_failedcount,
                        self._tcp_client_config.tcp_keepalive_probes_sendintervalms,
                        self)

                # Go
                self.current_socket.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPIDLE, self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000)
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPINTVL, self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000)
                self.current_socket.setsockopt(SOL_TCP, TCP_KEEPCNT, self._tcp_client_config.tcp_keepalive_probes_failedcount)

                # Check
                v = self.current_socket.getsockopt(SOL_SOCKET, SO_KEEPALIVE)
                if v != 1:
                    logger.warn("SO_KEEPALIVE mismatch, having=%s, required=1", v)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPIDLE)
                if v != self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000:
                    logger.warn("TCP_KEEPIDLE mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_senddelayms / 1000)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPINTVL)
                if v != self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000:
                    logger.warn("TCP_KEEPINTVL mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_sendintervalms / 1000)

                v = self.current_socket.getsockopt(SOL_TCP, TCP_KEEPCNT)
                if v != self._tcp_client_config.tcp_keepalive_probes_failedcount:
                    logger.warn("TCP_KEEPCNT mismatch, having=%s, required=%s", v, self._tcp_client_config.tcp_keepalive_probes_failedcount)

            # Non-blocking mode
            self.current_socket.setblocking(0)
            self.current_socket.settimeout(None)
            logger.debug("TcpSimpleClient : connect : non blocking mode set, self=%s", self)

            # Start the read/write loops
            self._read_greenlet = gevent.spawn(self._read_loop)
            self._write_greenlet = gevent.spawn(self._write_loop)
            logger.debug("TcpSimpleClient : connect : r/w loops started, self=%s", self)

            # Done
            logger.debug("TcpSimpleClient : connect : done, self=%s", self)
            return True

        except Exception as e:
            # Logs
            logger.error("TcpSimpleClient : connect : Exception, ex=%s, self=%s", SolBase.extostr(e), self)

            # Call disconnect
            self._callback_disconnect()

            # Exit
            return False

    # =================================
    # DISCONNECT
    # =================================

    def disconnect(self):
        """
        Disconnect from server. Return true upon success.
        :return True if success, false otherwise.
        """
        try:
            logger.debug("TcpSimpleClient : disconnect : entering, self=%s", self)

            # Check
            if not self.is_connected:
                logger.debug("TcpSimpleClient : disconnect : not connected, doing nothing, self=%s", self)
                return False

            # Disconnect
            if self.current_socket:
                logger.debug("TcpSimpleClient : disconnect : socket.shutdown, self=%s", self)
                # noinspection PyUnusedLocal
                try:
                    self.current_socket.shutdown(2)
                except Exception as e:
                    logger.debug("Exception on shutdown, ex=%s, self=%s", SolBase.extostr(e), self)
                    pass

                logger.debug("TcpSimpleClient : disconnect : socket.close, self=%s", self)
                self.current_socket.close()
                self.current_socket = None

            # Reset socket related context
            self._dt_created = SolBase.datecurrent()
            self._dt_last_recv = self._dt_created
            self._dt_last_send = self._dt_created
            self._disconnect_count.increment()
            self._local_addr = None
            self._local_port = None
            self._remote_addr = None
            self._remote_port = None

            # Signal
            logger.debug("TcpSimpleClient : disconnect : is_connected=False, self=%s", self)
            self.is_connected = False

            # Handle notifications at upper level
            go_call = False

            # Second level callback - Process in lock to avoid multiple notification for same simple client instance
            with self.on_disconnect_notifypoolcallbacklock:
                if self.on_disconnect_notifypoolcallbackcall and self.on_disconnect_notifypoolcallback:
                    logger.debug("TcpSimpleClient : disconnect : calling second level disconnect callback, callback=%s, self=%s", self.on_disconnect_notifypoolcallback, self)
                    self.on_disconnect_notifypoolcallbackcall = False
                    go_call = True

            # Out of lock : call
            # (ASYNC, because we may be called by read/write greenlet, which are going to be killed just after)
            if go_call:
                try:
                    # Call it synchronously (mantis 1907)
                    self.on_disconnect_notifypoolcallback(self)

                    # DO NOT MOVE THIS SLEEP, REDIS WILL DEADLOCK ON START
                    SolBase.sleep(0)
                except Exception as e:
                    logger.warn("on_disconnect_notifypoolcallback : call exception=%s", SolBase.extostr(e))

            # Greenlet reset after is_connected=False (will help to exit itself)
            if self._read_greenlet:
                logger.debug("TcpSimpleClient : disconnect : read kill, self=%s", self)
                self._read_greenlet.kill(GreenletExit, False)
                logger.debug("TcpSimpleClient : disconnect : read kill 1, self=%s", self)
                self._read_greenlet = None
                logger.debug("TcpSimpleClient : disconnect : read kill 2, self=%s", self)

            if self._write_greenlet:
                logger.debug("TcpSimpleClient : disconnect : write kill, self=%s", self)
                self._write_greenlet.kill(GreenletExit, False)
                logger.debug("TcpSimpleClient : disconnect : write kill 1, self=%s", self)
                self._write_greenlet = None
                logger.debug("TcpSimpleClient : disconnect : write kill 2, self=%s", self)

            logger.debug("TcpSimpleClient : disconnect : done, self=%s", self)
            return True

        except Exception as e:
            logger.error("TcpSimpleClient : disconnect : Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    # =================================
    # SSL HANDSHAKE
    # =================================

    def __do_ssl_handshake(self):
        """
        Do a ssl handshake. Method is using a greenlet, but is blocking for the caller.
        :return: Nothing.
        """

        # Go
        SolBase.sleep(0)
        ssl_greenlet = gevent.spawn(self.__do_ssl_handshake_internal)

        # Wait for complete
        SolBase.sleep(0)
        ssl_greenlet.join()

    def __do_ssl_handshake_internal(self):
        """
        Do a ssl handshake. Call _callback_disconnect() if error.
        :return: Nothing
        """
        try:
            logger.debug("__do_ssl_handshake_internal : start now, self=%s", self)
            SolBase.sleep(0)
            self.current_socket.do_handshake()
        except Exception as e:
            logger.error("__do_ssl_handshake_internal : exception=%s, self=%s", SolBase.extostr(e), self)
            self._callback_disconnect()
        finally:
            SolBase.sleep(0)

    # =================================
    # RECEIVE
    # =================================

    def _on_receive(self, binary_buffer):
        """
        Callback called upon server receive.
        - binary_buffer : a BINARY buffer received on the socket.

        !! CAUTION !!
        - IN ALL CASES THIS METHOD RECEIVE A BINARY BUFFER

        !! CAUTION !!
        - PYTHON 2.X / binary - str    => str is a byte array (or an ascii string) : nothing to do
        - PYTHON 2.X / text - unicode  => you receive a str, you may convert it to unicode (SolBase.binary_to_unicode)

        !! CAUTION !!
        - PYTHON 3.X / binary - bytes  => bytes is binary. You may convert it to a str using encoding.

                |  2.x                     |  3.x
        --------+--------------------------+-----------------------
        Bytes   |  'abc' <type 'str'>      |  b'abc' <type 'bytes'>
        Unicode | u'abc' <type 'unicode'>  |   'abc' <type 'str'>
        :param binary_buffer: Binary buffer received.
        :return: Nothing.
        """

        # Received something...
        logger.debug("TcpSimpleClient : _on_receive : binary_buffer=%s, self=%s", repr(binary_buffer), self)

        # Parse
        self._receive_current_buf = ProtocolParserTextDelimited.parse_protocol(self._receive_current_buf, binary_buffer, self._receive_queue, "\n")

    def get_recv_queue_len(self):
        """
        Get receive queue len
        :return The receive queue length.
        """
        return self._receive_queue.qsize()

    # =================================
    # RECEIVE QUEUE
    # =================================

    def get_from_receive_queue(self, block=False, timeout_sec=None):
        """
        Get a buffer from the receive queue.
        - If block is False, will return an item OR raise an Empty exception if no item.
        - If block is True AND timeOut=None, will wait forever for an item.
        - If block is True and timeout_sec>0, will wait for timeout_sec then raise Empty exception if no item.
        :param block: If true, will block.
        :param timeout_sec: The timeout in second.
        """

        return self._receive_queue.get(block, timeout_sec)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# Logger
import logging

from abc import ABCMeta, abstractmethod
from pysolbase.SolBase import SolBase

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpServerClientContextAbstractFactory(object):
    """
    ..  py:class:: TcpServerClientContextAbstractFactory()

    Abstract client context factory.
    """
    __metaclass__ = ABCMeta

    # noinspection PyUnusedLocal
    @abstractmethod
    def get_new_clientcontext(self, tcp_server, client_id, client_socket, client_addr):
        """
        ..  py:function:: get_new_clientcontext(tcp_server, client_id, client_socket, client_addr)

        Return a new client context instance.

        :param tcp_server: The tcpserver instance.
        :param client_id: an integer, which is the unique id of this client.
        :param client_socket: The server socket.
        :param client_addr: The remote addr information.
        :return Returned object MUST be a subclass of TcpServerClientContext.
        """
        raise Exception("get_new_clientcontext : this MUST be implemented")



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

import logging

from pysolbase.SolBase import SolBase

from pysoltcp.tcpserver.basefactory.TcpServerClientContextAbstractFactory import TcpServerClientContextAbstractFactory
from pysoltcp.tcpserver.clientcontext.TcpServerClientContext import TcpServerClientContext

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpServerClientContextFactory(TcpServerClientContextAbstractFactory):
    """
    Default factory.
    """

    # noinspection PyMethodMayBeStatic
    def get_new_clientcontext(self, tcp_server, client_id, client_socket, client_addr):
        """
        Return a new client context instance.
        :param tcp_server : The tcpserver instance.
        :param client_id : an integer, which is the unique id of this client.
        :param client_socket : The server socket.
        :param client_addr : The remote addr information.
        :return Returned object MUST be a subclass of TcpServerClientContext.
        """
        return TcpServerClientContext(tcp_server, client_id, client_socket, client_addr)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# Logger
import logging

from gevent.queue import Queue
from pysolbase.SolBase import SolBase

from pysoltcp.tcpbase.ProtocolParserTextDelimited import ProtocolParserTextDelimited
from pysoltcp.tcpserver.clientcontext.TcpServerClientContext import TcpServerClientContext

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpServerQueuedClientContext(TcpServerClientContext):
    """
    Tcp server client context.
    """

    def __init__(self, tcp_server, client_id, client_socket, client_addr):
        """
        Constructor.
        :param tcp_server: The tcp server.
        :param client_id: The client id.
        :param client_socket: The client socket.
        :param client_addr: The client remote address.
        :return Nothing
        """

        # Base - we provide two callback :
        # - one for disconnecting ourselves
        # - one to notify socket receive buffer
        TcpServerClientContext.__init__(self, tcp_server, client_id, client_socket, client_addr)

        # Receive queue
        self.__receive_queue = Queue()

        # Receive current buffer
        self.__receive_current_buf = None

    # ================================
    # TO STRING OVERWRITE
    # ================================

    def __str__(self):
        """
        To string override
        :return: A string
        :rtype string
        """

        return "q.recv.size={0}*{1}".format(
            self.__receive_queue.qsize(),
            TcpServerClientContext.__str__(self)
        )

    # ===============================
    # RECEIVE
    # ===============================

    def _on_receive(self, binary_buffer):
        """
        Called on socket receive. Method parse the protocol and put receive queue.
        :param binary_buffer: The binary buffer received.
        :return Nothing.
        """

        # Got something
        logger.debug("TcpServerQueuedClientContext : _on_receive called, binary_buffer=%s, self=%s", repr(binary_buffer), self)

        # Parse
        self.__receive_current_buf = ProtocolParserTextDelimited.parse_protocol(self.__receive_current_buf, binary_buffer, self.__receive_queue, "\n")

    def get_recv_queue_len(self):
        """
        Get receive queue len.
        :return An integer.
        """
        return self.__receive_queue.qsize()

    # =================================
    # RECEIVE QUEUE
    # =================================

    def get_from_receive_queue(self, block=False, timeout_sec=None):
        """
        Get a buffer from the receive queue.
        :param block: If True, will block.
        :param timeout_sec: Timeout in seconds.
        :return An item queued.
        - If block is False, will return an item OR raise an Empty exception if no item.
        - If block is True AND timeOut=None, will wait forever for an item.
        - If block is True and timeout_sec>0, will wait for timeout_sec then raise Empty exception if no item.
        """

        return self.__receive_queue.get(block, timeout_sec)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""
from distutils.core import setup

import re
from setuptools import find_packages


def requirement_read(req_file):
    """
    Doc
    :param req_file: Doc
    :return: Doc
    """
    req_list = list()
    for row_buffer in open(req_file).readlines():
        # Skip empty
        if len(row_buffer.strip()) == 0:
            continue
            # Skip "- ..."
        if re.match("^-", row_buffer):
            continue
            # Skip "# ..."
        if re.match("^#", row_buffer):
            continue

        # Ok
        req_list.append(row_buffer)
    return req_list


# ===========================
# SETUP
# ===========================

p_name = "pysoltcp"
p_author = "Laurent Champagnac"
p_email = "champagnac.laurent@gmail.com"
p_url = "https://knock.center"
p_version = "1.0.1"

setup(

    # Project details
    name=p_name,
    author=p_author,
    author_email=p_email,
    url=p_url,
    description=p_name + " APIs",

    # Version, format : Major.Minor.Revision
    version=p_version,

    # Packages
    packages=find_packages(exclude=["*_test*", "_*"]),
    include_package_data=True,

    # License & read me
    license=open("LICENSE.txt").read(),
    long_description=open("README.md").read(),

    # Data files
    data_files=[
        ("", ["requirements_test.txt", "requirements.txt", "README.md", "LICENSE.txt"]),
    ],

    # Classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Natural Language :: English",
    ],

    # Dependencies
    install_requires=requirement_read("requirements.txt"),

    # Dependencies : test
    tests_require=requirement_read("requirements_test.txt"),

    # Zip
    zip_safe=False,
)



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

# Logger
import logging

import gevent
from gevent.queue import Empty
from pysolbase.SolBase import SolBase
from pysolmeters.Meters import Meters

from pysoltcp.tcpbase.SignaledBuffer import SignaledBuffer
from pysoltcp.tcpbase.TcpSocketManager import TcpSocketManager

SolBase.logging_init()
logger = logging.getLogger(__name__)


class TcpServerClientContext(TcpSocketManager):
    """
    Tcp server client context.
    """

    def __init__(self, tcp_server, client_id, client_socket, client_addr, cb_stop_asynch=None, cb_on_receive=None):
        """
        Constructor.
        :param tcp_server: The tcp server instance.
        :param client_id: The client id.
        :param client_socket: The client socket.
        :param client_addr: The client remote address.
        :param cb_stop_asynch: Callback to call upon stop. If None, self.stop_asynch is used.
        :param cb_on_receive: Callback to call upon socket receive. If none, self._on_receive is used.
        :return Nothing.
        """

        # Check
        if client_socket is None:
            logger.error("client_socket is None")
            raise Exception("client_socket is None")
        elif client_addr is None:
            logger.error("client_addr is None")
            raise Exception("client_addr is None")
        elif client_id is None:
            logger.error("client_id is None")
            raise Exception("client_id is None")
        elif tcp_server is None:
            logger.error("tcp_server is None")
            raise Exception("tcp_server is None")

        # Base - we provide two callback :
        # - one for disconnecting ourselves
        # - one to notify socket receive buffer

        # Stop callback
        cur_cb_stop_asynch = cb_stop_asynch
        if cb_stop_asynch is None:
            # Use default
            cur_cb_stop_asynch = self.stop_asynch

        # Receive callback
        cur_cb_on_receive = cb_on_receive
        if cur_cb_on_receive is None:
            # Use default
            cur_cb_on_receive = self._on_receive

        TcpSocketManager.__init__(self, cur_cb_stop_asynch, cur_cb_on_receive)

        # Store
        self._client_id = client_id
        self._client_addr = client_addr

        # Base store
        self.current_socket = client_socket

        # Server store
        self._tcp_server = tcp_server

        # Greenlets
        self._read_greenlet = None
        self._write_greenlet = None
        self._control_greenlet = None

        # Faster str
        self._socket_local_ip = None
        self._socket_local_port = None
        self._socket_remote_ip = None
        self._socket_remote_port = None

    # ================================
    # TO STRING OVERWRITE
    # ================================

    def __get_socket_local_ip(self):
        """
        Get socket local ip
        :return: String
        :rtype: str
        """
        if self._socket_local_ip:
            return self._socket_local_ip
        # noinspection PyBroadException
        try:
            self._socket_local_ip = self.current_socket.getsockname()[0]
            return self._socket_local_ip
        except:
            return "exception"

    def __get_socket_local_port(self):
        """
        Get socket local ip
        :return: String
        :rtype: str
        """

        if self._socket_local_port:
            return self._socket_local_port
        # noinspection PyBroadException
        try:
            self._socket_local_port = self.current_socket.getsockname()[1]
            return self._socket_local_port
        except:
            return "exception"

    def __get_socket_remote_ip(self):
        """
        Get socket remote ip
        :return: String
        :rtype: str
        """
        if self._socket_remote_ip:
            return self._socket_remote_ip
        # noinspection PyBroadException
        try:
            self._socket_remote_ip = self.current_socket.getpeername()[0]
            return self._socket_remote_ip
        except:
            return "exception"

    def __get_socket_remote_port(self):
        """
        Get socket remote port
        :return: String
        :rtype: str
        """
        if self._socket_remote_port:
            return self._socket_remote_port

        # noinspection PyBroadException
        try:
            self._socket_remote_port = self.current_socket.getpeername()[1]
            return self._socket_remote_port
        except:
            return "exception"

    def __str__(self):
        """
        To string override
        :return: A string
        :rtype string
        """

        return "c.id={0}*c.cli={1}:{2}/{3}:{4}*{5}".format(
            self._client_id,
            self.__get_socket_local_ip(),
            self.__get_socket_local_port(),
            self.__get_socket_remote_ip(),
            self.__get_socket_remote_port(),
            TcpSocketManager.__str__(self)
        )

    # ===============================
    # START
    # ===============================

    def start(self):
        """
        Start processing our socket read/write.
        :return True if success, false otherwise.
        """
        try:
            # Check
            if self.is_connected:
                logger.warn("TcpServerClientContext : start : already connected, doing nothing, self=%s", self)
                return False

            # Done
            self.is_connected = True
            logger.debug("TcpServerClientContext : start : now connected, starting r/w loops, self=%s", self)

            # Start the read/write loops
            self._read_greenlet = gevent.spawn(self._read_loop)
            self._write_greenlet = gevent.spawn(self._write_loop)

            # Start the control greenlet
            self._schedule_control_greenlet()

            # Done
            logger.debug("TcpServerClientContext : start : done, self=%s", self)
            return True
        except Exception as e:
            logger.error("TcpServerClientContext : start : Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    # ===============================
    # STOP
    #
    # - stop_asynch
    # => used ONLY if stop comes from OUR instance (read/write loop initiated stop, due to socket issues, control stuff)
    # => This delegates the stop to the tcp_server, which will process it asynchronously, ending with a stop_synch call.
    #
    # - stop_synch
    # => used by external components to request a stop.
    # ===============================

    def stop_asynch(self):
        """
        Request a stop to the tcp_server.. It will call _stopInternal(), via tcp_server._remove_client_asynch.
        Do NOT OVERRIDE this method.
        :return True if success, false otherwise.
        """
        try:
            if not self.is_connected:
                return True
            else:
                # noinspection PyProtectedMember
                self._tcp_server._remove_client_asynch(self.get_client_id())
                return True
        except Exception as e:
            logger.error("TcpServerClientContext : stopFromServer : Exception, ex=%s, self=%s", SolBase.extostr(e), self)
            return False

    def stop_synch(self):
        """
        Stop processing our socket read/write.
        CAUTION : This method NOT be called if the server is stopping.
        Do NOT call this from ourselves, instead use stop_asynch, you may/will experience unexpected behaviors.
        This method is RESERVED for EXTERNAL class stop requests.
        You can OVERRIDE this method at HIGHER level to be informed of a socket closure.
        :return True if success, false otherwise.
        """

        return True

    def stop_synch_internal(self):
        """
        Stop processing our socket read/write.
        Reserved for PURE in-memory stop operations (greenlet stop, counter put mainly)
        NEVER, NEVER perform any kind of non-memory operations here.
        For instance, are FORDIDEN in higher level implementation of stop_synch_internal :
        - Any socket send/recv
        - Any external queries (redis/mongo, whatever)
        JUST HANDLE IN MEMORY STUFF.
        :return True if success, false otherwise.
        """

        try:
            logger.debug("TcpServerClientContext : disconnect : entering, self=%s", self)

            # Check
            if not self.is_connected:
                logger.debug("TcpServerClientContext : disconnect : not connected, doing nothing, self=%s", self)
                return False

            # Signal (move on top, try to avoid some TcpManager warn logs while stopping)
            self.is_connected = False

            # Timeout unschedule
            self._unschedule_ssl_handshake_timeout()

            # Control unschedule
            self._unschedule_control_greenlet()

            # Disconnect
            # Close the socket in this case (should not cover mantis 1173)
            TcpSocketManager.safe_close_socket(self.current_socket)
            self.current_socket = None

            # Greenlet reset after is_connected=False (will help to exit itself)
            if self._read_greenlet:
                self._read_greenlet.kill()
                self._read_greenlet = None

            if self._write_greenlet:
                self._write_greenlet.kill()
                self._write_greenlet = None

            # Flush out the send queue now, and decrement pending bytes to send
            total_len = 0
            while True:
                try:
                    item = self.send_queue.get(False)
                    if isinstance(item, str):
                        total_len += len(item)
                    elif isinstance(item, SignaledBuffer):
                        total_len += len(item.binary_buffer)
                except Empty:
                    break

            # Decrement
            logger.debug("TcpServerClientContext : disconnect : decrementing, total_len=%s", total_len)
            Meters.aii("tcp.server.server_bytes_send_pending", -total_len)

            # Over
            logger.debug("TcpServerClientContext : disconnect : done, self=%s", self)
            return True

        except Exception as e:
            logger.error("TcpServerClientContext : disconnect : Exception, ex=%s, self=%s", SolBase.extostr(e),
                         self)
            return False
        finally:
            # Session duration stats
            sec = SolBase.datediff(self._dt_created) / 1000
            Meters.dtci("tcp.server.session_duration_second", sec)

    # ===============================
    # GETTER / SETTER
    # ===============================

    def get_client_id(self):
        """
        Getter
        :return The client id.
        """
        return self._client_id

    def get_client_socket(self):
        """
        Getter
        :return The client socket.
        """
        return self.current_socket

    def get_client_addr(self):
        """
        Getter
        return: The client remote address.
        """
        return self._client_addr

    # ===============================
    # RECEIVE
    # ===============================

    def _on_receive(self, binary_buffer):
        """
        Called on socket receive.
        :param binary_buffer: The received buffer.
        :return Nothing.
        """

        # Got something
        logger.debug("TcpServerClientContext : _on_receive called, binary_buffer=%s, self=%s", repr(binary_buffer), self)
        pass

    # ===============================
    # CONTROL
    # ===============================

    def _schedule_control_greenlet(self):
        """
        Schedule the control greenlet for next run.
        Note: We do not use lock to minimize per socket memory usage...
        :return: Nothing.
        """

        # Check
        if not self.is_connected:
            return

        # Get
        ms = self._tcp_server.get_effective_controlinterval_ms()
        if ms <= 0:
            return

        # Schedule it !
        self._control_greenlet = gevent.spawn_later(ms * 0.001, self._control_us)

    def _unschedule_control_greenlet(self):
        """
        unschedule the control greenlet.
        Note: We do not use lock to minimize per socket memory usage...
        :return: Nothing.
        """

        if self._control_greenlet:
            self._control_greenlet.kill(block=False)
            self._control_greenlet = None

    def _control_us(self):
        """
        Control us. May force ourself to exit upon inactivity.
        Note: We do not use lock to minimize per socket memory usage...
        :return: Nothing.
        """

        logger.info("Entering")

        # Check
        if not self.is_connected:
            return

        # Get settings
        c = self._tcp_server.get_tcpserver_config()

        # Check absolute
        absolute_ms = c.socket_absolute_timeout_ms
        if absolute_ms > 0:
            running_ms = SolBase.datediff(self._dt_created)
            if running_ms > absolute_ms:
                logger.debug("Absolute reached, running_ms=%s, absolute_ms=%s, self=%s", running_ms, absolute_ms, self)
                # Kill ourself if we are running
                if self.is_connected:
                    self.stop_asynch()
                return
            else:
                logger.debug("Absolute not reached, running_ms=%s, absolute_ms=%s, self=%s", running_ms, absolute_ms,
                             self)
                pass

        # Check relative
        relative_ms = c.socket_relative_timeout_ms
        if relative_ms > 0:
            # Get last receive ms
            last_recv_ms = SolBase.datediff(self._dt_last_recv)

            # Get last send ms
            last_send_ms = SolBase.datediff(self._dt_last_send)

            # We keep the minimum of both
            last_ms = min(last_recv_ms, last_send_ms)

            # Check
            if last_ms > relative_ms:
                # Kill ourself
                logger.debug("Relative reached, last_recv_ms=%s, last_send_ms=%s, last_ms=%s, relative_ms=%s, self=%s", last_recv_ms, last_send_ms, last_ms, relative_ms, self)
                if self.is_connected:
                    self.stop_asynch()
                return
            else:
                logger.debug("Relative not reached, last_recv_ms=%s, last_send_ms=%s, last_ms=%s, relative_ms=%s, self=%s", last_recv_ms, last_send_ms, last_ms, relative_ms, self)

        # Schedule next check
        self._schedule_control_greenlet()



"""
# -*- coding: utf-8 -*-
# ===============================================================================
#
# Copyright (C) 2013/2017 Laurent Labatut / Laurent Champagnac
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
# ===============================================================================
"""

from gevent.event import Event


class SignaledBuffer(object):
    """
    Send class for tcp, with signal upon send completion.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.__buffer = None
        self.__send_event = Event()

    def __set_buffer(self, local_buffer):
        """
        Set the binary local_buffer.
        :param local_buffer: The binary local_buffer.
        :type local_buffer: str
        """
        self.__buffer = local_buffer

    def __get_buffer(self):
        """
        Getter
        :return: The binary localBuffer
        :rtype: str
        """
        return self.__buffer

    def __get_event(self):
        """
        Get the event.
        :return: Event
        :rtype: Event
        """
        return self.__send_event

    binary_buffer = property(__get_buffer, __set_buffer)
    send_event = property(__get_event)
